unit xeRegistryIO;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, ComCtrls, Buttons, ImgList, ExtCtrls, XmlIni;

type
  TfrmRegistryIO = class(TForm)
    tvKeyTreeView: TTreeView;
    lbSections: TListBox;
    mmValue: TRichEdit;
    lbKeys: TListBox;
    Label1: TLabel;
    Label2: TLabel;
    Label3: TLabel;
    Label4: TLabel;
    btnOK: TBitBtn;
    btnCancel: TBitBtn;
    edEditKeyPath: TEdit;
    cbEditRootKey: TComboBox;
    btnRefresh: TBitBtn;
    btnDeleteKey: TBitBtn;
    ImageList1: TImageList;
    lvKeys: TListView;
    Label5: TLabel;
    Label6: TLabel;
    Label7: TLabel;
    Bevel1: TBevel;
    btnPreview: TBitBtn;
    procedure FormCreate(Sender: TObject);
    procedure tvKeyTreeViewClick(Sender: TObject);
    procedure tvKeyTreeViewExpanding(Sender: TObject; Node: TTreeNode;
      var AllowExpansion: Boolean);
    procedure edEditKeyPathExit(Sender: TObject);
    procedure cbEditRootKeyChange(Sender: TObject);
    procedure btnRefreshClick(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure lbSectionsClick(Sender: TObject);
    procedure lbKeysClick(Sender: TObject);
    procedure btnPreviewClick(Sender: TObject);
  private
    FAllowCreateRegKey: Boolean;
    FEditRootKey: HKEY;
    FEditKeyPath: string;
    FXmlIni: TXmlIni;

    procedure PrepareSubKeys(Node: TTreeNode);

    function NodeToKeyPath(Node: TTreeNode): string;

    procedure ReloadXISections;
    procedure ReloadXIKeys;
    procedure ReloadXIKey;

    function GetEditRootKey: HKEY;
    procedure SetEditRootKey(Value: HKEY);
    function GetEditKeyPath: string;
    procedure SetEditKeyPath(Value: string);
  public
    function ShowBox: Boolean;

    property AllowCreateRegKey: Boolean read FAllowCreateRegKey write FAllowCreateRegKey;
    property EditRootKey: HKEY read GetEditRootKey write SetEditRootKey;
    property EditKeyPath: string read GetEditKeyPath write SetEditKeyPath;
  end;

var
  frmRegistryIO: TfrmRegistryIO;

implementation

{$R *.dfm}

uses
  Registry;

type
  TRootKeys = (
      rkHKEY_CLASSES_ROOT, rkHKEY_CURRENT_USER, rkHKEY_LOCAL_MACHINE, rkHKEY_USERS,
      rkHKEY_PERFORMANCE_DATA, rkHKEY_CURRENT_CONFIG, rkHKEY_DYN_DATA);
const
  RootKeyNames: array [TRootKeys] of string = (
      'HKEY_CLASSES_ROOT', 'HKEY_CURRENT_USER', 'HKEY_LOCAL_MACHINE', 'HKEY_USERS',
      'HKEY_PERFORMANCE_DATA', 'HKEY_CURRENT_CONFIG', 'HKEY_DYN_DATA');
  RootKeysHKEYMap: array [TRootKeys] of HKEY = (
      HKEY_CLASSES_ROOT, HKEY_CURRENT_USER, HKEY_LOCAL_MACHINE, HKEY_USERS,
      HKEY_PERFORMANCE_DATA, HKEY_CURRENT_CONFIG, HKEY_DYN_DATA);

{ TfrmRegistryIO }

procedure TfrmRegistryIO.FormCreate(Sender: TObject);
var
  k: TRootKeys;
begin
  cbEditRootKey.Items.Clear;
  for k := Low(TRootKeys) to High(TRootKeys) do
    cbEditRootKey.Items.Add(RootKeyNames[k]);

  FXmlIni := TXmlIni.Create;

  AllowCreateRegKey := False;
  EditRootKey := HKEY_LOCAL_MACHINE;
  EditKeyPath := '\';
end;

procedure TfrmRegistryIO.FormDestroy(Sender: TObject);
begin
  FXmlIni.Free;
end;

function TfrmRegistryIO.ShowBox: Boolean;
begin
  btnRefreshClick(nil);
  Result := ShowModal = mrOK;
end;

procedure TfrmRegistryIO.PrepareSubKeys(Node: TTreeNode);
var
  Reg: TRegistry;
  KeySL: TStringList;
  i: Integer;
  function HasSubKey(KeyPath: string): Boolean;
  begin
    Result := False;
    
    if Reg.OpenKey(KeyPath, False) then
    begin
      try
        Result := Reg.HasSubKeys;
      finally
        Reg.CloseKey;
      end;
    end;
  end;
var
  KeyPath: string;
  KeyNode: TTreeNode;
begin
  if (Node <> nil) and (not Node.HasChildren or (Node.Count > 0)) then
    Exit;

  Reg := TRegistry.Create;
  KeySL := TStringList.Create;
  try
    Reg.RootKey := EditRootKey;

    if Node = nil then
    begin
      tvKeyTreeView.Items.Clear;
      tvKeyTreeView.Items.Add(nil, cbEditRootKey.Items[cbEditRootKey.ItemIndex]);

      KeyPath := '\';

      if HasSubKey(KeyPath) then
      begin
        tvKeyTreeView.Items.GetFirstNode.HasChildren := True;
        PrepareSubKeys(tvKeyTreeView.Items.GetFirstNode);
        tvKeyTreeView.Items.GetFirstNode.Expanded := True;
      end;
    end
    else
    begin
      KeyPath := NodeToKeyPath(Node);

      if Reg.OpenKey(KeyPath, False) then
      begin
        try
          Reg.GetKeyNames(KeySL);
        finally
          Reg.CloseKey;
        end;
      end;
      
      for i := 0 to KeySL.Count-1 do
      begin
        KeyNode := tvKeyTreeView.Items.AddChild(Node, KeySL[i]);
        KeyNode.HasChildren := HasSubKey(KeyPath + KeySL[i] + '\');
      end;
    end;
  finally
    KeySL.Free;
    Reg.Free;
  end;
end;

function TfrmRegistryIO.NodeToKeyPath(Node: TTreeNode): string;
begin
  if (Node = nil) or (Node = tvKeyTreeView.Items.GetFirstNode) then
    Result := '\'
  else
  begin
    Result := Node.Text + '\';
    if Node.Parent <> nil then
      Result := NodeToKeyPath(Node.Parent) + Result;
  end;
end;

procedure TfrmRegistryIO.ReloadXISections;
var
  i: Integer;
  OldSection: string;
begin
  FXmlIni.LoadFromRegistry(FEditRootKey, FEditKeyPath);

  lbSections.Color := clWindow;
  lbKeys.Color := clWindow;
  mmValue.Color := clWindow;

  OldSection := '';
  if lbSections.ItemIndex >= 0 then
    OldSection := lbSections.Items[lbSections.ItemIndex];

  FXmlIni.Collation;
  lbSections.Items.Clear;
  for i := 0 to FXmlIni.Count-1 do
    lbSections.Items.Add(FXmlIni.SectionByIndex[i].Name);
  lbSections.Sorted := True;

  if (OldSection <> '') and (lbSections.Items.IndexOf(OldSection) >= 0) then
    lbSections.ItemIndex := lbSections.Items.IndexOf(OldSection);

  ReloadXIKeys;
end;

procedure TfrmRegistryIO.ReloadXIKeys;
var
  i: Integer;
  OldKey, SelSection: string;
begin
  lbKeys.Items.Clear;

  if lbSections.Count > 0 then
  begin
    if lbSections.ItemIndex < 0 then
      lbSections.ItemIndex := 0;

    SelSection := lbSections.Items[lbSections.ItemIndex];

    OldKey := '';
    if lbKeys.ItemIndex >= 0 then
      OldKey := lbKeys.Items[lbKeys.ItemIndex];

    FXmlIni[SelSection].Collation;
    for i := 0 to FXmlIni[SelSection].Count-1 do
      lbKeys.Items.Add(FXmlIni[SelSection].KeyByIndex[i].Name);
    lbKeys.Sorted := True;

    if (OldKey <> '') and (lbKeys.Items.IndexOf(OldKey) >= 0) then
      lbKeys.ItemIndex := lbKeys.Items.IndexOf(OldKey);
  end;

  ReloadXIKey;
end;

procedure TfrmRegistryIO.ReloadXIKey;
begin
  mmValue.Lines.Text := '';
  
  if (lbSections.Items.Count > 0) and (lbKeys.Items.Count > 0) then
  begin
    if lbSections.ItemIndex < 0 then
      lbSections.ItemIndex := 0;
      
    if lbKeys.ItemIndex < 0 then
      lbKeys.ItemIndex := 0;
      
    if FXmlIni.Exists(lbSections.Items[lbSections.ItemIndex], lbKeys.Items[lbKeys.ItemIndex]) then
      mmValue.Lines.Text := FXmlIni[lbSections.Items[lbSections.ItemIndex]][lbKeys.Items[lbKeys.ItemIndex]];
  end;
end;

function TfrmRegistryIO.GetEditRootKey: HKEY;
begin
  Result := FEditRootKey;
end;

procedure TfrmRegistryIO.SetEditRootKey(Value: HKEY);
var
  i: TRootKeys;
begin
  if FEditRootKey <> Value then
  begin
    FEditRootKey := Value;

    for i := Low(TRootKeys) to High(TRootKeys) do
      if FEditRootKey = RootKeysHKEYMap[i] then
      begin
        cbEditRootKey.ItemIndex := Integer(i);
        Break;
      end;

    tvKeyTreeView.Items.Clear;
    PrepareSubKeys(nil);

    EditKeyPath := '\';
  end;
end;

function TfrmRegistryIO.GetEditKeyPath: string;
begin
  SetEditKeyPath(edEditKeyPath.Text);
  Result := edEditKeyPath.Text;
end;

procedure TfrmRegistryIO.SetEditKeyPath(Value: string);
  function GetSelNode(Node: TTreeNode; KeyPath: string): TTreeNode;
  var
    P, i: Integer;
    FindNode: string;
    Reg: TRegistry;
  begin
    Result := Node;

    if KeyPath <> '' then
    begin
      PrepareSubKeys(Node);

      P := Pos('\', KeyPath);
      if P > 0 then
      begin
        FindNode := Copy(KeyPath, 1, P-1);
        for i := 0 to Node.Count-1 do
          if UpperCase(Node.Item[i].Text) = UpperCase(FindNode) then
          begin
            Result := GetSelNode(Node.Item[i], Copy(KeyPath, P+1, MaxInt));
            Exit;
          end;

        if Showing then
        begin
          if FAllowCreateRegKey then
          begin
            if MessageDlg('Key "' + Copy(FEditKeyPath, 1, Length(FEditKeyPath) - Length(KeyPath)) + FindNode +
                '" does not exists. Do you want to create key ?',
                mtWarning, [mbYes, mbNo], 0) = mrYes then
            begin
              Reg := TRegistry.Create;
              try
                Reg.RootKey := EditRootKey;

                if Reg.OpenKey(Copy(FEditKeyPath, 1, Length(FEditKeyPath) - Length(KeyPath)) + FindNode, True) then
                  Reg.CloseKey;
              finally
                Reg.Free;
              end;

              Node.DeleteChildren;
              Node.HasChildren := True;

              Result := GetSelNode(Node, KeyPath);
            end;
          end
          else
            MessageDlg('Key "' + Copy(FEditKeyPath, 1, Length(FEditKeyPath) - Length(KeyPath)) + FindNode +
                '" does not exists.', mtError, [mbOK], 0);
        end;
      end;
    end;
  end;
var
  Reg: TRegistry;
  ValueNameList: TStringList;
  RegDataInfo: TRegDataInfo;
  ValueText: string;
  i: Integer;
begin
  if FEditKeyPath <> Value then
  begin
    FEditKeyPath := Value;

    if (FEditKeyPath = '') or (FEditKeyPath[1] <> '\') then
      FEditKeyPath := '\' + FEditKeyPath;

    if FEditKeyPath[Length(FEditKeyPath)] <> '\' then
      FEditKeyPath := FEditKeyPath + '\';

    tvKeyTreeView.Selected := GetSelNode(tvKeyTreeView.Items.GetFirstNode, Copy(FEditKeyPath, 2, MaxInt));

    Reg := TRegistry.Create;
    ValueNameList := TStringList.Create;
    try
      lvKeys.Items.Clear;

      Reg.RootKey := GetEditRootKey;

      if Reg.OpenKey(FEditKeyPath, False) then
      begin
        Reg.GetValueNames(ValueNameList);
        for i := 0 to ValueNameList.Count-1 do
          with lvKeys.Items.Add do
          begin
            Caption := ValueNameList[i];
            
            Reg.GetDataInfo(ValueNameList[i], RegDataInfo);

            case RegDataInfo.RegData of
              rdString, rdExpandString:
                ValueText := Reg.ReadString(ValueNameList[i]);
              rdInteger:
                ValueText := IntToStr(Reg.ReadInteger(ValueNameList[i]));
              rdUnknown, rdBinary:
                ValueText := '<binary>';
            end;
            SubItems.Add(ValueText);
          end;
      end;
    finally
      ValueNameList.Free;
      Reg.Free;
    end;

    edEditKeyPath.Text := FEditKeyPath;

    FXmlIni.Clear;

    lbSections.Items.Clear;
    lbKeys.Items.Clear;
    mmValue.Lines.Clear;

    lbSections.Color := clSilver;
    lbKeys.Color := clSilver;
    mmValue.Color := clSilver;
  end;
end;

procedure TfrmRegistryIO.tvKeyTreeViewClick(Sender: TObject);
begin
  EditKeyPath := NodeToKeyPath(tvKeyTreeView.Selected);
end;

procedure TfrmRegistryIO.tvKeyTreeViewExpanding(Sender: TObject;
  Node: TTreeNode; var AllowExpansion: Boolean);
begin
  PrepareSubKeys(Node);
end;

procedure TfrmRegistryIO.cbEditRootKeyChange(Sender: TObject);
begin
  EditRootKey := RootKeysHKEYMap[TRootKeys(cbEditRootKey.ItemIndex)];
end;

procedure TfrmRegistryIO.edEditKeyPathExit(Sender: TObject);
begin
  EditKeyPath := edEditKeyPath.Text;
end;

procedure TfrmRegistryIO.btnRefreshClick(Sender: TObject);
var
  OldEditKeyPath: string;
begin
  OldEditKeyPath := EditKeyPath;
  tvKeyTreeView.Items.Clear;
  PrepareSubKeys(nil);
  FEditKeyPath := '\';
  EditKeyPath := OldEditKeyPath;
end;

procedure TfrmRegistryIO.lbSectionsClick(Sender: TObject);
begin
  ReloadXIKeys;
end;

procedure TfrmRegistryIO.lbKeysClick(Sender: TObject);
begin
  ReloadXIKey;
end;

procedure TfrmRegistryIO.btnPreviewClick(Sender: TObject);
begin
  ReloadXISections;
end;

end.
